/*
** client.c -- a stream socket client demo
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <arpa/inet.h>

#include <string>
#include <iostream>
#include <fstream>
#include <vector>

#include <pcl/point_cloud.h>
#include <pcl/point_types.h>
#include <pcl/visualization/cloud_viewer.h>

#define PORT "3490" // the port client will be connecting to 

#define MAXDATASIZE 100 // max number of bytes we can get at once 

using namespace std;

// get sockaddr, IPv4 or IPv6:
void *get_in_addr(struct sockaddr *sa)
{
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*)sa)->sin_addr);
    }

    return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

int main(int argc, char *argv[])
{
    int sockfd, numbytes;  
    char buf[MAXDATASIZE];
    struct addrinfo hints, *servinfo, *p;
    int rv;
    char s[INET6_ADDRSTRLEN];

    if (argc != 2) {
        fprintf(stderr,"usage: client hostname\n");
        exit(1);
    }

    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    if ((rv = getaddrinfo(argv[1], PORT, &hints, &servinfo)) != 0) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
        return 1;
    }

    // loop through all the results and connect to the first we can
    for(p = servinfo; p != NULL; p = p->ai_next) {
        if ((sockfd = socket(p->ai_family, p->ai_socktype,
                p->ai_protocol)) == -1) {
            perror("client: socket");
            continue;
        }

        if (connect(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
            close(sockfd);
            perror("client: connect");
            continue;
        }

        break;
    }

    if (p == NULL) {
        fprintf(stderr, "client: failed to connect\n");
        return 2;
    }

    inet_ntop(p->ai_family, get_in_addr((struct sockaddr *)p->ai_addr),
            s, sizeof s);
    printf("client: connecting to %s\n", s);

    freeaddrinfo(servinfo); // all done with this structure

    //if ((numbytes = recv(sockfd, buf, MAXDATASIZE-1, 0)) == -1){perror("recv"); exit(1);}
    //buf[numbytes] = '\0';
    //printf("client: received '%s'\n",buf);
    
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ>); //empty point cloud pointer
    //string extraction = "-70.0 -50.0 40.0 60.0 -10.0 10.0 0.01";
    //write( sockfd, extraction.c_str(), extraction.c_str().size() );
    //write extraction info
    cout << "Enter extraction details (site,xstart,xend,ystart,yend,zstart,zend,resolution): " << endl;
    string site;
    float x_start, x_end, y_start, y_end, z_start, z_end; //gede-pillar nice region
    float resolution;
    cin >> site >> x_start >> x_end >> y_start >> y_end >> z_start >> z_end >> resolution;
    
    write( sockfd, reinterpret_cast<char*>( &x_start ), sizeof x_start );
    write( sockfd, reinterpret_cast<char*>( &x_end ), sizeof x_end );
    write( sockfd, reinterpret_cast<char*>( &y_start ), sizeof y_start );
    write( sockfd, reinterpret_cast<char*>( &y_end ), sizeof y_end );
    write( sockfd, reinterpret_cast<char*>( &z_start ), sizeof z_start );
    write( sockfd, reinterpret_cast<char*>( &z_end ), sizeof z_end );
    write( sockfd, reinterpret_cast<char*>( &resolution ), sizeof resolution );
    send(sockfd, site.c_str(), site.length(), 0);


    while(1)
    {
        float x,y,z;
        read( sockfd, reinterpret_cast<char*>( &x ), sizeof x ); //break?
        read( sockfd, reinterpret_cast<char*>( &y ), sizeof y );
        read( sockfd, reinterpret_cast<char*>( &z ), sizeof z );
        if(x==0 && y==0 && z==0) {cout << "NOWDONE" << endl;break;}
        pcl::PointXYZ newpoint(x,y,z);
		cloud->push_back (newpoint);
    }
    close(sockfd);
    cout << cloud->width << endl;
    
    pcl::visualization::CloudViewer viewer ("Simple Cloud Viewer");
	viewer.showCloud (cloud);
	while (!viewer.wasStopped ())
	{
	}

    return 0;
}
