#include <pcl/point_cloud.h>
#include <pcl/octree/octree.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/visualization/cloud_viewer.h>
#include <pcl/octree/octree_iterator.h>
#include <pcl/octree/octree_impl.h>
#include <pcl/octree/octree_nodes.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <ctime>

using namespace std;
int main (int argc, char** argv)
{
	//start timer
	struct timeval startTime, endTime; long diff;
	gettimeofday(&startTime, NULL);
	string indexlocation = "/media/Storage/jordanbinindex/"; //location of the index

	//read in cloud size and bounding box
	ifstream fin((indexlocation+"index.txt").c_str());
	long cloud_size;
	fin >> cloud_size;
	float min_x, max_x, min_y, max_y, min_z, max_z;
	fin >> min_x >> max_x >> min_y >> max_y >> min_z >> max_z;
	
	float a,b,c,d,e,f;
	string filename;
	int numpoints;
	while(fin >> a >> b >> c >> d >> e >> f)
	{
		fin >> filename; fin >> numpoints;
		if(numpoints>0)
		{
			ifstream binaryin( (indexlocation+filename).c_str(), ios::binary);
			pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ>);
			cloud->width = numpoints; cloud->height = 1;
			cloud->points.resize (cloud->width * cloud->height);
			for (size_t i = 0; i < cloud->points.size (); ++i)
			{
				float x,y,z;
				binaryin.read( reinterpret_cast<char*>( &x ), sizeof x );
				binaryin.read( reinterpret_cast<char*>( &y ), sizeof y );
				binaryin.read( reinterpret_cast<char*>( &z ), sizeof z );
				cloud->points[i].x = x; cloud->points[i].y = y; cloud->points[i].z = z;
			}
			binaryin.close();
			//create octree
			float resolution = 0.1f;
			pcl::octree::OctreePointCloud<pcl::PointXYZ> octree (resolution);
			octree.setInputCloud (cloud);
			octree.addPointsFromInputCloud ();
			//cout << "Octree depth: " << octree.getTreeDepth() << endl;
			
			//get data
			pcl::octree::OctreePointCloud< pcl::PointXYZ >::LeafNodeIterator itLeafs ( octree );
			vector<vector<int> > pointIndexes;
			int max_size = 0;
			while (*++itLeafs)
			{
				vector<int> temp;
				itLeafs.getData(temp);
				if(temp.size() > max_size) max_size = temp.size();
				pointIndexes.push_back(temp);
			}
			//write out data
			ofstream binout((indexlocation+filename+".new").c_str(), ios::binary);
			
			vector<int> numbers; numbers.resize(max_size);
			for(int i=0; i<max_size; i++)numbers[i]=i;
			int random[max_size];
			srand ( time(NULL) );
			for(int i=0; i<max_size; i++)
			{
				int randomnumber = rand()%numbers.size();
				random[i]=numbers[randomnumber];
				numbers[randomnumber]=numbers.back();
				numbers.pop_back();
			}
			for(int position=0; position<max_size; position++)
			{
				for(int i=0; i<pointIndexes.size(); i++)
				{
					if(random[position] < pointIndexes[i].size())
					{
						float x = cloud->points[pointIndexes[i][random[position]]].x;
						float y = cloud->points[pointIndexes[i][random[position]]].y;
						float z = cloud->points[pointIndexes[i][random[position]]].z;
						binout.write( reinterpret_cast<char*>( &x ), sizeof x );
						binout.write( reinterpret_cast<char*>( &y ), sizeof y );
						binout.write( reinterpret_cast<char*>( &z ), sizeof z );
					}
				}
			}
			//cout << "Highest Index: " << position << endl;
			binout.close();
		}
	} 
	
	gettimeofday(&endTime, NULL);
	diff =  (endTime.tv_sec - startTime.tv_sec) * 1000000L;
	diff += (endTime.tv_usec - startTime.tv_usec);
	cout << "Total time taken: " << (diff / 1000L) << " ms" << endl;
}
