#include <pcl/point_cloud.h>
#include <pcl/kdtree/kdtree_flann.h>
#include <pcl/io/pcd_io.h>
#include <pcl/point_types.h>
#include <pcl/visualization/cloud_viewer.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <ctime>

int
main (int argc, char** argv)
{
  struct timeval startTime, endTime; long diff;
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ>);

  // Parse first lines of file
  gettimeofday(&startTime, NULL);
  std::ifstream fin("../../s1.ptx");
  int num_cols, points_per_col;
  fin >> num_cols >> points_per_col;
  std::string skipline;
  for(int i=0; i<8; i++) getline(fin,skipline); //skip 8 lines
  fin >> std::ws; //clear any whitespace at end of line
  
  cloud->width = num_cols*points_per_col;
  cloud->height = 1;
  cloud->points.resize (cloud->width * cloud->height);
  std::cout << "Cloud contains " << cloud->width * cloud->height << " points" << std::endl;
  
  float intensity = 0;
  for (size_t i = 0; i < cloud->points.size (); ++i)
  {
    fin >> cloud->points[i].x;
    fin >> cloud->points[i].y;
    fin >> cloud->points[i].z;
    fin >> intensity;
  }
  gettimeofday(&endTime, NULL);
  diff =  (endTime.tv_sec - startTime.tv_sec) * 1000000L;
  diff += (endTime.tv_usec - startTime.tv_usec);
  std::cout << "Time to read input and build cloud: " << (diff / 1000L) << " ms" << std::endl;
  
  gettimeofday(&startTime, NULL);
  float resolution = 128.0f;

  pcl::KdTreeFLANN<pcl::PointXYZ> kdtree;
  kdtree.setInputCloud (cloud);
  
  gettimeofday(&endTime, NULL);
  diff =  (endTime.tv_sec - startTime.tv_sec) * 1000000L;
  diff += (endTime.tv_usec - startTime.tv_usec);
  std::cout << "Time to build kdtree: " << (diff / 1000L) << " ms" << std::endl;

  pcl::PointXYZ searchPoint;
  

  int K = 5000000;
  //searchPoint.x = 0.001938; searchPoint.y = 0.641983; searchPoint.z = -1.334427;
  //searchPoint.x=-2.314377; searchPoint.y=5.118454; searchPoint.z=-1.176376;
  //searchPoint.x=-0.551193; searchPoint.y=1.218704; searchPoint.z=-1.312057; //hole
  searchPoint.x=3.34619; searchPoint.y=-0.465078; searchPoint.z=2.17752; //arch
  //searchPoint.x=-13.6074; searchPoint.y=-16.9; searchPoint.z=10.3; K=1000000;//tree

   // K nearest neighbor search
  gettimeofday(&startTime, NULL);
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud2 (new pcl::PointCloud<pcl::PointXYZ>);
  cloud2->width = K;
  cloud2->height = 1;
  cloud2->points.resize (cloud2->width * cloud2->height);

  std::vector<int> pointIdxNKNSearch;
  std::vector<float> pointNKNSquaredDistance;

  std::cout << "K nearest neighbor search at (" << searchPoint.x 
            << " " << searchPoint.y 
            << " " << searchPoint.z
            << ") with K=" << K << std::endl;

  if (kdtree.nearestKSearch (searchPoint, K, pointIdxNKNSearch, pointNKNSquaredDistance) > 0)
  {
    for (size_t i = 0; i < pointIdxNKNSearch.size (); ++i)
    {
    	cloud2->points[i].x = cloud->points[ pointIdxNKNSearch[i] ].x;
    	cloud2->points[i].y = cloud->points[ pointIdxNKNSearch[i] ].y;
    	cloud2->points[i].z = cloud->points[ pointIdxNKNSearch[i] ].z;
    }
  }
  gettimeofday(&endTime, NULL);
  diff =  (endTime.tv_sec - startTime.tv_sec) * 1000000L;
  diff += (endTime.tv_usec - startTime.tv_usec);
  std::cout << "Time to perform nearest neighbour search: " << (diff / 1000L) << " ms" << std::endl;

  //WRITE OUT POINT CLOUD TO PCD FILE
  //pcl::io::savePCDFileASCII ("test_pcd.pcd", *cloud);
  //std::cerr << "Saved " << cloud->points.size () << " data points to test_pcd.pcd." << std::endl;
  //... populate cloud
  
  pcl::visualization::CloudViewer viewer ("Simple Cloud Viewer");
  viewer.showCloud (cloud2);
  while (!viewer.wasStopped ())
  {
  }

}
