/*
** server.c -- a stream socket server demo
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>

#include <string>
#include <iostream>
#include <fstream>
#include <vector>

#define PORT "3490"  // the port users will be connecting to

#define BACKLOG 10     // how many pending connections queue will hold

using namespace std;

void extraction(int new_fd, float x_start, float x_end, float y_start, float y_end, float z_start, float z_end, float resolution, string site)
{
    cout << "Performing extraction: " << site << " " << x_start << " " << x_end << " " << y_start << " " << y_end << " " << z_start << " " << z_end << " " << resolution << endl;
    string path;
    if(site=="gede")path="/media/Storage/gedebinindex/"; //location of the index you wish to load
    else if(site=="mosque")path="/media/Storage/mosque/";
	ifstream fin((path+"index.txt").c_str()); //open the root index file
	//read in cloud size and bounding box
	long cloud_size;
	fin >> cloud_size;
	float min_x, max_x, min_y, max_y, min_z, max_z;
	fin >> min_x >> max_x >> min_y >> max_y >> min_z >> max_z;

	vector<ifstream*> partial; vector<int> partial_size;
	vector<ifstream*> entire; vector<int> entire_size;
	//the desired resolution of the region as a fraction of the total number of available points
	//float resolution = 0.01;	 
	//user set region for extraction
	//float x_start=-70.0, x_end=-50.0, y_start=40.0, y_end=60.0, z_start=-10.0, z_end=10.0; //gede-pillar nice region
		
	float a,b,c,d,e,f;
	string filename;
	int numpoints;
	while(fin >> a >> b >> c >> d >> e >> f)
	{
		fin >> filename;
		fin >> numpoints;
		if(numpoints>0 && x_start <= a && a < x_end && y_start <= c && c < y_end && z_start <= e && e < z_end)
		{
			if(b<x_end && d<y_end && f<z_end)
			{
				
				entire.push_back( new ifstream((path+filename).c_str(), ios::binary) );
				entire_size.push_back( numpoints );
			}
			else
			{
				partial.push_back( new ifstream((path+filename).c_str(), ios::binary) );
				partial_size.push_back( numpoints);
			}
		}
	}
	fin.close();
	int numsent;		
	for(int i=0; i<entire.size(); i++)
	{
		float x,y,z;
		int numadded = 0; int numneeded = resolution*entire_size[i];
		while(true)
		{
			entire[i]->read( reinterpret_cast<char*>( &x ), sizeof x );
			if(entire[i]->eof() || numadded>=numneeded)break;
			entire[i]->read( reinterpret_cast<char*>( &y ), sizeof y );
    		entire[i]->read( reinterpret_cast<char*>( &z ), sizeof z );
    		//write point out
    		if(x==0 && y==0 && z==0)continue;
    		write( new_fd, reinterpret_cast<char*>( &x ), sizeof x );
			write( new_fd, reinterpret_cast<char*>( &y ), sizeof x );
			write( new_fd, reinterpret_cast<char*>( &z ), sizeof x );
			numsent++;
			numadded++;
		}
		entire[i]->close();
	}
	for(int i=0; i<partial.size(); i++)
	{
		float x,y,z;
		int numadded = 0; int numneeded = resolution*entire_size[i];
		while(true)
		{
			partial[i]->read( reinterpret_cast<char*>( &x ), sizeof x );
			if(partial[i]->eof() || numadded>=numneeded)break;
    		partial[i]->read( reinterpret_cast<char*>( &y ), sizeof y );
    		partial[i]->read( reinterpret_cast<char*>( &z ), sizeof z );
			if(x_start<=x && x<=x_end && y_start<=y && y<=y_end && z_start<=z && z<=z_end)
			{
				//write point out
				if(x==0 && y==0 && z==0)continue;
				write( new_fd, reinterpret_cast<char*>( &x ), sizeof x );
				write( new_fd, reinterpret_cast<char*>( &y ), sizeof x );
				write( new_fd, reinterpret_cast<char*>( &z ), sizeof x );
				numsent++;
				numadded++;
			}
		}
		partial[i]->close();
	}
	cout << numsent << endl;
	printf("NOW DONE");
	float x=0,y=0,z=0;
	write( new_fd, reinterpret_cast<char*>( &x ), sizeof x );
	write( new_fd, reinterpret_cast<char*>( &y ), sizeof x );
	write( new_fd, reinterpret_cast<char*>( &z ), sizeof x );
}


void sigchld_handler(int s)
{
    while(waitpid(-1, NULL, WNOHANG) > 0);
}

// get sockaddr, IPv4 or IPv6:
void *get_in_addr(struct sockaddr *sa)
{
    if (sa->sa_family == AF_INET) {
        return &(((struct sockaddr_in*)sa)->sin_addr);
    }

    return &(((struct sockaddr_in6*)sa)->sin6_addr);
}
int main(void)
{
    int sockfd, new_fd;  // listen on sock_fd, new connection on new_fd
    struct addrinfo hints, *servinfo, *p;
    struct sockaddr_storage their_addr; // connector's address information
    socklen_t sin_size;
    struct sigaction sa;
    int yes=1;
    char s[INET6_ADDRSTRLEN];
    int rv;

    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE; // use my IP

    if ((rv = getaddrinfo(NULL, PORT, &hints, &servinfo)) != 0) {
        fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
        return 1;
    }

    // loop through all the results and bind to the first we can
    for(p = servinfo; p != NULL; p = p->ai_next) {
        if ((sockfd = socket(p->ai_family, p->ai_socktype,
                p->ai_protocol)) == -1) {
            perror("server: socket");
            continue;
        }

        if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes,
                sizeof(int)) == -1) {
            perror("setsockopt");
            exit(1);
        }

        if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
            close(sockfd);
            perror("server: bind");
            continue;
        }

        break;
    }

    if (p == NULL)  {
        fprintf(stderr, "server: failed to bind\n");
        return 2;
    }

    freeaddrinfo(servinfo); // all done with this structure

    if (listen(sockfd, BACKLOG) == -1) {
        perror("listen");
        exit(1);
    }

    sa.sa_handler = sigchld_handler; // reap all dead processes
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = SA_RESTART;
    if (sigaction(SIGCHLD, &sa, NULL) == -1) {
        perror("sigaction");
        exit(1);
    }

    printf("server: waiting for connections...\n");

    while(1) {  // main accept() loop
        sin_size = sizeof their_addr;
        new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
        if (new_fd == -1) {
            perror("accept");
            continue;
        }

        inet_ntop(their_addr.ss_family,
            get_in_addr((struct sockaddr *)&their_addr),
            s, sizeof s);
        printf("server: got connection from %s\n", s);

        if (!fork()) { // this is the child process
            close(sockfd); // child doesn't need the listener
            //read the dimension for extraction
            float x_start, x_end, y_start, y_end, z_start, z_end, resolution;
            read( new_fd, reinterpret_cast<char*>( &x_start ), sizeof x_start );
            read( new_fd, reinterpret_cast<char*>( &x_end ), sizeof x_end );
            read( new_fd, reinterpret_cast<char*>( &y_start ), sizeof y_start );
            read( new_fd, reinterpret_cast<char*>( &y_end ), sizeof y_end );
            read( new_fd, reinterpret_cast<char*>( &z_start ), sizeof z_start );
            read( new_fd, reinterpret_cast<char*>( &z_end ), sizeof z_end );
            read( new_fd, reinterpret_cast<char*>( &resolution ), sizeof resolution );
            //read the site
            char site[200];
            int length = recv(new_fd, site, 200-1, 0);
            site[length] = '\0';
            string strsite(site);
            
                        
            
            extraction(new_fd,x_start, x_end, y_start, y_end, z_start, z_end, resolution,strsite);
            close(new_fd);
            exit(0);
        }
        close(new_fd);  // parent doesn't need this
    }

    return 0;
}
